/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2011 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.io.IOException;
import java.util.Calendar;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import com.bitctrl.Constants;
import com.bitctrl.util.Interval;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.iav.gllib.gllib.Ganglinie;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.modell.db.DbGanglinieDaten;
import de.bsvrz.iav.gllib.gllib.modell.db.GlSpeicher;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellAutomatischesLernenEreignis;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.kalender.Ereigniskalender;
import de.bsvrz.sys.funclib.bitctrl.kalender.KalenderAnfrage;
import de.bsvrz.sys.funclib.bitctrl.kalender.KalenderEvent;
import de.bsvrz.sys.funclib.bitctrl.kalender.KalenderListener;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.EreignisTyp;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.onlinedaten.OdEreignisKalenderAntwort;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.parameter.PdEreignisTypParameter;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.NetzBestandTeil;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Sammelt und kapselt die Ausgangsdaten für das Lernen von Ganglinien.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Ausgangsdaten {

	/**
	 * Sortiert absteigend nach der Priorität der Ereignistypen.
	 */
	private class EreignisTypNachPrioritaet implements Comparator<EreignisTyp> {

		@Override
		public int compare(final EreignisTyp t1, final EreignisTyp t2) {
			final long p1, p2;

			assert parameterEreignisTyp.get(t1).isValid();
			assert parameterEreignisTyp.get(t2).isValid();

			p1 = parameterEreignisTyp.get(t1).getPrioritaet();
			p2 = parameterEreignisTyp.get(t2).getPrioritaet();
			if (p1 > p2) {
				return -1;
			} else if (p1 < p2) {
				return 1;
			}
			return 0;
		}

	}

	/**
	 * Empfänger für die Antwort des Ereigniskalender.
	 */
	private class KalenderEmpfaenger implements KalenderListener {

		/**
		 * Extrahiert die Ereignistypen der Ereignisse im Lernzyklus.
		 */
		@Override
		public void antwortEingetroffen(final KalenderEvent e) {
			if (!e.getAbsenderZeichen().equals(absenderZeichen)) {
				// Antwort ist für eine andere Anfrage bestimmt.
				return;
			}

			// Wieder abmelden
			kalender.removeKalenderListener(kalenderEmpfaenger);

			for (final OdEreignisKalenderAntwort.Daten.Zustandswechsel z : e
					.getZustandswechsel()) {
				final EreignisTyp typ;
				PdEreignisTypParameter.Daten paramEreignisTyp;
				PdGanglinienModellAutomatischesLernenEreignis.Daten paramEreignisTypLernen;

				if (((z.getZeitstempel() <= lernzeitraum.getStart())
						&& !z.isVerkehrlichGueltig())
						|| ((z.getZeitstempel() >= lernzeitraum.getEnd())
								&& z.isVerkehrlichGueltig())) {
					// Ein endendes Ereignis zu Beginn des Lernzeitraum und ein
					// beginnendes Ereignis am Ende des Lernzeitraum wird
					// ignoriert.
					continue;
				}

				typ = z.getEreignis().getEreignisTyp();
				paramEreignisTyp = typ
						.getParameterDatensatz(PdEreignisTypParameter.class)
						.abrufenDatum();
				paramEreignisTypLernen = typ.getParameterDatensatz(
						PdGanglinienModellAutomatischesLernenEreignis.class)
						.abrufenDatum();
				assert paramEreignisTyp.isValid() : "Der Ereignistyp " + typ //$NON-NLS-1$
						+ " muss parametriert sein."; //$NON-NLS-1$
				assert paramEreignisTypLernen
						.isValid() : "Der Lernparameter für Ereignistyp " //$NON-NLS-1$
								+ typ + " muss parametriert sein."; //$NON-NLS-1$
				if (paramEreignisTyp.isValid()
						&& paramEreignisTypLernen.isValid()) {
					parameterEreignisTyp.put(typ, paramEreignisTyp);
					parameterEreignisTypLernen.put(typ, paramEreignisTypLernen);
					ereignisTypen.add(typ);
				} else if (!paramEreignisTyp.isValid()) {
					LogTools.log(log,
							new BetriebsmeldungDaten(getMq().getSystemObject()),
							GlLernenMsg.WarningParameterUngueltig, getMq(), typ,
							paramEreignisTyp);
				} else if (paramEreignisTypLernen.isValid()) {
					LogTools.log(log,
							new BetriebsmeldungDaten(getMq().getSystemObject()),
							GlLernenMsg.WarningParameterUngueltig, getMq(), typ,
							paramEreignisTypLernen);
				}
			}

			EreignisTyp[] hilfe;
			String txt;
			hilfe = ereignisTypen
					.toArray(new EreignisTyp[ereignisTypen.size()]);
			txt = new String();
			for (int i = 0; i < hilfe.length; ++i) {
				EreignisTyp typ;

				typ = hilfe[i];
				txt += typ + " (Priorität " //$NON-NLS-1$
						+ parameterEreignisTyp.get(typ).getPrioritaet() + ")"; //$NON-NLS-1$
				if (i < (ereignisTypen.size() - 1)) {
					txt += ", "; //$NON-NLS-1$
				}
			}
			assert txt.length() > 0 : "Es müssen Ereignisse anliegen."; //$NON-NLS-1$
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.FineAnliegendeEreignisse, mq, txt);

			// Bescheid sagen, dass Kalenderantwort eingetroffen ist
			synchronized (mutex) {
				mutex.notify();
			}
		}

	}

	/** Die Verbindung zum Datenverteiler. */
	private final ClientDavInterface dav;

	/** Der Logger der Klasse. */
	private final Debug log = Debug.getLogger();

	/** Die Verbindung zum Ereigniskalender. */
	private final Ereigniskalender kalender;

	/** Der Empfänger für die Kalenderantwort. */
	private final KalenderEmpfaenger kalenderEmpfaenger;

	/** Der Messquerschnitt für den gelernt wird. */
	private final MessQuerschnittAllgemein mq;

	/** Der gesamte Lernzeitraum, kann mehrere Tage umfassen. */
	private final Interval lernzeitraum;

	/**
	 * Ganglinie, die aus den Archivdaten bestimmt wird und gelernt werden soll
	 * (fuer das Approximationsverfahren B-Spline).
	 */
	private GanglinieMQ archivGanglinieBSpline;

	/**
	 * Ganglinie, die aus den Archivdaten bestimmt wird und gelernt werden soll
	 * (fuer das Approximationsverfahren CubicSpline).
	 */
	private GanglinieMQ archivGanglinieCubicSpline;

	/**
	 * Ganglinie, die aus den Archivdaten bestimmt wird und gelernt werden soll
	 * (fuer das Approximationsverfahren Polyline).
	 */
	private GanglinieMQ archivGangliniePolyline;

	/** Cached die Archivganglinien pro Aspekt. */
	private final Map<Aspect, GanglinieMQ> archivGanglinienCache = new HashMap<>();

	/** Die historischen Ganglinien am Messquerschnitt. */
	private DbGanglinieDaten historischeGanglinien;

	/** Synchronisiert die Kalenderantwort. */
	private final Object mutex = new Object();

	/**
	 * Die Ereignistypen im aktuellen Lernzeitraum, absteigend sortiert nach
	 * Priorität.
	 */
	private final SortedSet<EreignisTyp> ereignisTypen;

	/** Die Parameter der Ereignistypen im aktuellen Lernzeitraum. */
	private final Map<EreignisTyp, PdEreignisTypParameter.Daten> parameterEreignisTyp;

	/** Die Lernparameter der Ereignistypen im aktuellem Lernzeitraum. */
	private final Map<EreignisTyp, PdGanglinienModellAutomatischesLernenEreignis.Daten> parameterEreignisTypLernen;

	/** Diese Kennung wird für Anfragen an den Kalender verwendet. */
	private final String absenderZeichen;

	/**
	 * Initialisierung den Lernprozess.
	 *
	 * @param mq
	 *            der Messquerschnitt für den die Ganglinien gelernt werden.
	 * @param lernzeitraum
	 *            der Zeitraum in dem gelernt werden soll.
	 */
	public Ausgangsdaten(final MessQuerschnittAllgemein mq,
			final Interval lernzeitraum) {
		this.mq = mq;
		absenderZeichen = "GlLernen" + mq + ", " + lernzeitraum; //$NON-NLS-1$//$NON-NLS-2$

		this.lernzeitraum = lernzeitraum;
		assert lernzeitraum
				.getLength() == Constants.MILLIS_PER_DAY : "Der Lernzyklus muss genau einen Tag umfassen."; //$NON-NLS-1$

		dav = ObjektFactory.getInstanz().getVerbindung();
		kalenderEmpfaenger = new KalenderEmpfaenger();
		kalender = Ereigniskalender.getInstanz();
		kalender.addKalenderListener(kalenderEmpfaenger);

		ereignisTypen = new TreeSet<>(new EreignisTypNachPrioritaet());
		parameterEreignisTyp = new HashMap<>();
		parameterEreignisTypLernen = new HashMap<>();
	}

	/**
	 * Ruft vom Datenverteiler die zum Lernen notwendigen Daten ab. Dazu gehören
	 * die Archivdaten und die Ereignisse im Lernzeitraum.
	 *
	 * @return {@code false}, wenn die verfügbaren Daten zum Lernen nicht
	 *         ausreichen.
	 */
	public boolean abrufen() {
		final long offsetBSpline, offsetCubicSpline, offsetPolyline;
		final Calendar calBSpline, calCubicSpline, calPolyline;
		final EreignisTyp[] hilfe;
		String txt;

		// Im Lernzeitraum anliegende Ereignisse abfragen
		starteKalenderanfrage();

		// Historische Ganglinien bestimmen
		historischeGanglinien = GlSpeicher.getInstanz().read(getMq().getPid());

		/**
		 * Vorher:
		 */
		// historischeGanglinien =
		// mq.getParameterDatensatz(PdGanglinie.class).abrufenDatum();
		LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
				GlLernenMsg.FinestGesicherteHistorischeGanglinien, mq,
				historischeGanglinien);

		// Archivganglinie des Lernzeitraum bestimmen
		archivGanglinienCache.clear();
		archivGanglinieBSpline = bestimmeArchivGanglinie(
				GanglinieMQ.APPROX_BSPLINE);
		archivGanglinieCubicSpline = bestimmeArchivGanglinie(
				GanglinieMQ.APPROX_CUBICSPLINE);
		archivGangliniePolyline = bestimmeArchivGanglinie(
				GanglinieMQ.APPROX_POLYLINE);
		if ((archivGanglinieBSpline == null)
				|| (archivGanglinieCubicSpline == null)
				|| (archivGangliniePolyline == null)) {
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.WarningKeineArchivdaten, mq);
			return false;
		}

		// absolute Zeitstempel der Analyseganglinie in relative übersetzen
		calBSpline = Calendar.getInstance();
		calBSpline.setTimeInMillis(archivGanglinieBSpline.getStuetzstellen()
				.get(0).getZeitstempel());
		calBSpline.set(Calendar.HOUR_OF_DAY, 0);
		calBSpline.set(Calendar.MINUTE, 0);
		calBSpline.set(Calendar.SECOND, 0);
		calBSpline.set(Calendar.MILLISECOND, 0);
		offsetBSpline = -calBSpline.getTimeInMillis();
		GanglinienMQOperationen.verschiebe(archivGanglinieBSpline,
				offsetBSpline);
		LogTools.log(log, GlLernenMsg.FinerNormierteArchivganglinie,
				archivGanglinieBSpline.getMessQuerschnitt(),
				archivGanglinieBSpline);

		calCubicSpline = Calendar.getInstance();
		calCubicSpline.setTimeInMillis(archivGanglinieCubicSpline
				.getStuetzstellen().get(0).getZeitstempel());
		calCubicSpline.set(Calendar.HOUR_OF_DAY, 0);
		calCubicSpline.set(Calendar.MINUTE, 0);
		calCubicSpline.set(Calendar.SECOND, 0);
		calCubicSpline.set(Calendar.MILLISECOND, 0);
		offsetCubicSpline = -calCubicSpline.getTimeInMillis();
		GanglinienMQOperationen.verschiebe(archivGanglinieCubicSpline,
				offsetCubicSpline);
		LogTools.log(log, GlLernenMsg.FinerNormierteArchivganglinie,
				archivGanglinieCubicSpline.getMessQuerschnitt(),
				archivGanglinieCubicSpline);

		calPolyline = Calendar.getInstance();
		calPolyline.setTimeInMillis(archivGangliniePolyline.getStuetzstellen()
				.get(0).getZeitstempel());
		calPolyline.set(Calendar.HOUR_OF_DAY, 0);
		calPolyline.set(Calendar.MINUTE, 0);
		calPolyline.set(Calendar.SECOND, 0);
		calPolyline.set(Calendar.MILLISECOND, 0);
		offsetPolyline = -calPolyline.getTimeInMillis();
		GanglinienMQOperationen.verschiebe(archivGangliniePolyline,
				offsetPolyline);
		LogTools.log(log, GlLernenMsg.FinerNormierteArchivganglinie,
				archivGangliniePolyline.getMessQuerschnitt(),
				archivGangliniePolyline);

		// Warte bis Kalender geantwortet hat
		synchronized (mutex) {
			if (ereignisTypen.isEmpty()) {
				// Gewartet werden muss nur, wenn noch keine Ereignis da sind.

				try {
					mutex.wait(5 * Constants.MILLIS_PER_MINUTE);
				} catch (final InterruptedException ex) {
					LogTools.log(log,
							new BetriebsmeldungDaten(mq.getSystemObject()),
							GlLernenMsg.WarningInterrupt, mq);
					return false;
				}
			}
		}

		if (ereignisTypen.isEmpty()) {
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.WarningKeineEreignisse, mq);
			return false;
		}

		dav.sleep(5 * Constants.MILLIS_PER_SECOND);

		hilfe = ereignisTypen.toArray(new EreignisTyp[ereignisTypen.size()]);
		txt = new String();
		for (int i = 0; i < hilfe.length; ++i) {
			EreignisTyp typ;

			typ = hilfe[i];
			txt += typ + " (Priorität " //$NON-NLS-1$
					+ parameterEreignisTyp.get(typ).getPrioritaet() + ")"; //$NON-NLS-1$
			if (i < (ereignisTypen.size() - 1)) {
				txt += ", "; //$NON-NLS-1$
			}
		}
		assert txt.length() > 0 : "Es müssen Ereignisse anliegen."; //$NON-NLS-1$
		LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
				GlLernenMsg.FineAnliegendeEreignisse, mq, txt);

		return true;
	}

	/**
	 * Fragt den Ereigniskalender nach Ereignissen im Lernzyklus.
	 */
	private void starteKalenderanfrage() {
		final KalenderAnfrage anfrage;

		if (!kalender.isBereit()) {
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.ErrorKalenderNichtVerfuegbar, mq);
			return;
		}

		anfrage = new KalenderAnfrage();
		anfrage.setIntervall(lernzeitraum);
		if (mq.getStrassenSegment() instanceof NetzBestandTeil) {
			anfrage.getRaeumlicheGueltigkeit()
					.add((NetzBestandTeil) mq.getStrassenSegment());
		}
		try {
			kalender.sendeAnfrage(absenderZeichen, anfrage);
		} catch (final DatensendeException ex) {
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.ErrorBeiKalenderanfrageSenden, mq,
					ex.getLocalizedMessage());
			return;
		}
	}

	/**
	 * Ruft die Kurzzeitdaten im aktuellen Lernintervall aus dem Archiv ab. Gibt
	 * {@code null} zurück, wenn keine Daten im Archiv verfügbar sind.
	 *
	 * @param verfahren
	 *            das Approximationsverfahren.<br>
	 *            z.B.: {@link GanglinieMQ#APPROX_BSPLINE},
	 *            {@link GanglinieMQ#APPROX_CUBICSPLINE},
	 *            {@link GanglinieMQ#APPROX_POLYLINE},
	 *            {@link GanglinieMQ#APPROX_UNBESTIMMT}
	 * @return die Ganglinie die aus den Archivdaten gebildet wurde oder
	 *         {@code null}.
	 */
	private GanglinieMQ bestimmeArchivGanglinie(final Integer verfahren) {
		final ArchiveRequestManager archiv;
		final ArchiveDataKindCombination dataKinds;
		final DataDescription dbs;
		final ArchiveTimeSpecification timeSpec;
		final ArchiveDataSpecification spec;
		final ArchiveDataQueryResult antwort;
		final AttributeGroup atg;
		final Aspect asp;
		final Ganglinie<Double> gQKfz, gQLkw, gVPkw, gVLkw;
		final GanglinieMQ ganglinie;

		archiv = dav.getArchive();
		if (!archiv.isArchiveAvailable()) {
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.ErrorArchivNichtVerfuegbar, mq);
			return null;
		}

		// Anfrage zusammenstellen
		atg = dav.getDataModel()
				.getAttributeGroup("atg.verkehrsDatenKurzZeitMq"); //$NON-NLS-1$
		asp = GanglinienAspekte.getAspect(verfahren);

		// Cache pruefen:
		final GanglinieMQ cacheGanglinie = archivGanglinienCache.get(asp);
		if (cacheGanglinie != null) {
			return cacheGanglinie.clone();
		}

		dbs = new DataDescription(atg, asp);
		timeSpec = new ArchiveTimeSpecification(TimingType.DATA_TIME, false,
				lernzeitraum.getStart() - 1,
				lernzeitraum.getEnd() + Constants.MILLIS_PER_DAY);
		dataKinds = new ArchiveDataKindCombination(ArchiveDataKind.ONLINE);
		spec = new ArchiveDataSpecification(timeSpec, dataKinds,
				ArchiveOrder.BY_DATA_TIME, ArchiveRequestOption.NORMAL, dbs,
				mq.getSystemObject());

		// Anfrage verschicken und auf Antwort warten
		antwort = archiv.request(ArchiveQueryPriority.MEDIUM, spec);

		gQKfz = new Ganglinie<>();
		gQLkw = new Ganglinie<>();
		gVPkw = new Ganglinie<>();
		gVLkw = new Ganglinie<>();
		try {
			final ArchiveDataStream strom;
			ArchiveData archivdaten;

			assert antwort
					.getStreams().length == 1 : "Es darf nur einen Archivstream geben."; //$NON-NLS-1$

			// Daten auslesen und Ganglinie daraus erstellen
			strom = antwort.getStreams()[0];
			while ((archivdaten = strom.take()) != null) {
				final DataState status;

				status = archivdaten.getDataType();
				if (!status.equals(DataState.DATA)
						&& !status.equals(DataState.NO_DATA)
						&& !status.equals(DataState.NO_SOURCE)) {
					// Statuswerte des Archiv werden nicht in Ganglinie
					// übernommen
					continue;
				}

				final Data daten;
				final long zeitstempel;
				final Double qKfz;
				final Double qLkw;
				final Double vLkw;
				final Double vPkw;

				zeitstempel = archivdaten.getDataTime();
				daten = archivdaten.getData();
				if (daten == null) {
					// Definitionslücke
					qKfz = null;
					qLkw = null;
					vLkw = null;
					vPkw = null;
				} else {
					if (daten.getItem("QKfz").getScaledValue("Wert") //$NON-NLS-1$//$NON-NLS-2$
							.isState()) {
						// Definitionslücke
						qKfz = null;
					} else {
						qKfz = daten.getItem("QKfz").getScaledValue("Wert") //$NON-NLS-1$ //$NON-NLS-2$
								.doubleValue();
					}
					if (daten.getItem("QLkw").getScaledValue("Wert") //$NON-NLS-1$//$NON-NLS-2$
							.isState()) {
						// Definitionslücke
						qLkw = null;
					} else {
						qLkw = daten.getItem("QLkw").getScaledValue("Wert") //$NON-NLS-1$ //$NON-NLS-2$
								.doubleValue();
					}
					if (daten.getItem("VLkw").getScaledValue("Wert") //$NON-NLS-1$//$NON-NLS-2$
							.isState()) {
						// Definitionslücke
						vLkw = null;
					} else {
						vLkw = daten.getItem("VLkw").getScaledValue("Wert") //$NON-NLS-1$ //$NON-NLS-2$
								.doubleValue();
					}
					if (daten.getItem("VPkw").getScaledValue("Wert") //$NON-NLS-1$//$NON-NLS-2$
							.isState()) {
						// Definitionslücke
						vPkw = null;
					} else {
						vPkw = daten.getItem("VPkw").getScaledValue("Wert") //$NON-NLS-1$ //$NON-NLS-2$
								.doubleValue();
					}
				}
				gQKfz.put(zeitstempel, qKfz);
				gQLkw.put(zeitstempel, qLkw);
				gVPkw.put(zeitstempel, vPkw);
				gVLkw.put(zeitstempel, vLkw);
			}

			ganglinie = GanglinienMQOperationen.zusammenfuehren(gQKfz, gQLkw,
					gVPkw, gVLkw);
			ganglinie.setMessQuerschnitt(mq);
			if (ganglinie.size() >= 2) {
				GanglinienMQOperationen.verschiebeUmHalbesIntervall(ganglinie);
			}
			if (ganglinie.size() > 0) {
				GanglinienMQOperationen.auschneiden(ganglinie, lernzeitraum);
				LogTools.log(log,
						new BetriebsmeldungDaten(mq.getSystemObject()),
						GlLernenMsg.FineArchivganglinie, mq,
						GanglinienMQOperationen.formatierterText(ganglinie));

				archivGanglinienCache.put(asp, ganglinie);

				return ganglinie;
			}
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.InfoKeineArchivdaten, mq);
		} catch (final IllegalStateException ex) {
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.ErrorBeiArchivanfrage, mq,
					ex.getLocalizedMessage());
		} catch (final InterruptedException ex) {
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.ErrorBeiArchivanfrage, mq,
					ex.getLocalizedMessage());
		} catch (final IOException ex) {
			LogTools.log(log, new BetriebsmeldungDaten(mq.getSystemObject()),
					GlLernenMsg.ErrorBeiArchivanfrage, mq,
					ex.getLocalizedMessage());
		}

		return null;
	}

	/**
	 * Gibt den Messquerschnitt zurück, für den die Daten abgerufen werden.
	 *
	 * @return der Messquerschnitt.
	 */
	public MessQuerschnittAllgemein getMq() {
		return mq;
	}

	/**
	 * Gibt den Lernzetraum zurück, für den die Daten abgerufen werden.
	 *
	 * @return der Lernzeitraum.
	 */
	public Interval getLernzeitraum() {
		return lernzeitraum;
	}

	/**
	 * Gibt die Archivdaten im Lernzeitraum als Ganglinie zurück.
	 *
	 * @param verfahren
	 *            das Approximationsverfahren.<br>
	 *            z.B.: {@link GanglinieMQ#APPROX_BSPLINE},
	 *            {@link GanglinieMQ#APPROX_CUBICSPLINE},
	 *            {@link GanglinieMQ#APPROX_POLYLINE},
	 *            {@link GanglinieMQ#APPROX_UNBESTIMMT}
	 * @return die Archivganglinie fuer B-Spline.
	 */
	public GanglinieMQ getArchivGanglinie(final Integer verfahren) {
		GanglinieMQ gl = null;
		switch (verfahren) {
		case GanglinieMQ.APPROX_BSPLINE:
			gl = archivGanglinieBSpline;
			break;
		case GanglinieMQ.APPROX_CUBICSPLINE:
			gl = archivGanglinieCubicSpline;
			break;
		case GanglinieMQ.APPROX_POLYLINE:
			gl = archivGangliniePolyline;
			break;
		}
		return gl;
	}

	/**
	 * Gibt die historischen Ganglinien am zu lernenden Messquerschnit zurück.
	 *
	 * @return die historischen Ganglinien.
	 */
	public DbGanglinieDaten getHistorischeGanglinien() {
		return historischeGanglinien;
	}

	/**
	 * Gibt die Ereignistypen im Lernzeitraum absteigend sortiert nach Priorität
	 * zurück.
	 *
	 * @return die anliegenden Ereignistypen.
	 */
	public SortedSet<EreignisTyp> getEreignisTypen() {
		return ereignisTypen;
	}

	/**
	 * Gibt die Parameter der anliegenden Ereignistypen zurück.
	 *
	 * @return die Ereignistypparameter.
	 */
	public Map<EreignisTyp, PdEreignisTypParameter.Daten> getParamEreignisTyp() {
		return parameterEreignisTyp;
	}

	/**
	 * Gibt die Lernparameter der anliegenden Ereignistypen zurück.
	 *
	 * @return die Lernparameter der Ereignisse.
	 */
	public Map<EreignisTyp, PdGanglinienModellAutomatischesLernenEreignis.Daten> getParamEreignisTypLernen() {
		return parameterEreignisTypLernen;
	}

}
