/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.util.ResourceBundle;
import java.util.logging.Level;

import de.bsvrz.sys.funclib.bitctrl.daf.LogNachricht;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;

/**
 * Versorgt die Applikation mit lokalisierten Benutzernachrichten.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public enum GlLernenMsg implements LogNachricht {

	/** Nicht für Log-Ausgaben vorgesehen. */
	Applikationsname(null),

	/** Nicht für Log-Ausgaben vorgesehen. */
	ApplikationsnameKurz(null),

	/** Parameter: 1) Attributgruppe, 2) Parameterdaten. */
	IllegalStateExceptionLernParameterUngueltig(null),

	/** Parameter: 1) Aspekt. */
	IllegalArgumentExceptionAspektUngueltig(null),

	/** Parameter: 1) Thread, 2) Stack Trace. */
	ErrorUnbekannterFehler(Debug.ERROR),

	/** Parameter: 1) Parameterdaten. */
	ErrorUngueltigerLernParameter(Debug.ERROR),

	/** Parameter: 1) Messquerschnitt. */
	ErrorArchivNichtVerfuegbar(Debug.ERROR),

	/** Parameter: 1) Messquerschnitt, 2) Fehlermeldung. */
	ErrorBeiKalenderanfrageSenden(Debug.ERROR),

	/** Parameter: 1) Messquerschnitt, 2) Fehlermeldung. */
	ErrorBeiArchivanfrage(Debug.ERROR),

	/** Parameter: 1) Messquerschnitt. */
	ErrorKalenderNichtVerfuegbar(Debug.ERROR),

	/** Parameter: 1) Messquerschnitt. */
	WarningKeineArchivdaten(Debug.WARNING),

	/** Parameter: 1) Messquerschnitt. */
	WarningKeineEreignisse(Debug.WARNING),

	/** Parameter: 1) Messquerschnitt. */
	WarningKeinEreignistyp(Debug.WARNING),

	/** Parameter: 1) Messquerschnitt, 2) Ereignistyp. */
	WarningTriggerUnplausibel(Debug.WARNING),

	/** Parameter: 1) Messquerschnitt. */
	WarningInterrupt(Debug.WARNING),

	/** Parameter: 1) Messquerschnitt, 2) Systemobjekt, 3) Parameter. */
	WarningParameterUngueltig(Debug.WARNING),

	/**
	 * Parameter: 1) Messquerschnitt, 2) aktueller Abstand, 3) Erlaubter
	 * Abstand.
	 */
	WarningAbstandUeberschritten(Debug.WARNING),

	/** Parameter: 1) Messquerschnitt. */
	InfoKeineArchivdaten(Debug.INFO),

	/** Parameter: 1) Messquerschnitt. */
	InfoLernenAbgebrochen(Debug.INFO),

	/** Parameter: 1) Messquerschnitt. */
	InfoLernenErfolgreich(Debug.INFO),

	/** Ohne Parameter. */
	InfoGlLernenBereit(Debug.INFO),

	/**
	 * Parameter: 1) Messquerschnitt, 2) Index, 3) Anzahl Messquerschnitte.
	 */
	InfoLernenStartMQ(Debug.INFO),

	/** Ohne Parameter. */
	InfoLernenStartetWieder(Debug.INFO),

	/** Parameter: 1) Messquerschnitt, 2) Lernintervall. */
	InfoLernenStart(Debug.INFO),

	/** Parameter: 1) Messquerschnitt. */
	InfoLerneAbsoluteGanglinieStart(Debug.INFO),

	/** Parameter: 1) Messquerschnitt. */
	InfoLerneRelativeGanglinieStart(Debug.INFO),

	/** Keine Parameter. */
	InfoLernenNichtNotwendig(Debug.INFO),

	/** Ohne Parameter. */
	ConfigVerwaltungBereit(Debug.CONFIG),

	/** Parameter: 1) Anzahl gleichzeitig lernender Messquerschnitte. */
	ConfigLernenInitialisiert(Debug.CONFIG),

	/** Parameter: 1) Parameterdaten. */
	InfoParameterAenderung(Debug.INFO),

	/** Parameter: 1) Messquerschnitt, 2) Archivganglinie. */
	FineArchivganglinie(Debug.FINE),

	/** Parameter: 1) Messquerschnitt, 2) Ereignisse. */
	FineAnliegendeEreignisse(Debug.FINE),

	/** Parameter: 1) Messquerschnitt. */
	FineNeueGanglinie(Debug.FINE),

	/** Parameter: 1) Messquerschnitt. */
	FineAbstandUeberschritten(Debug.FINE),

	/** Parameter: 1) Messquerschnitt, 2) Gewicht. */
	FineVerschmelzeGanglinie(Debug.FINE),

	/**
	 * Parameter: 1) Messquerschnitt, 2) Ereignistyp, 3) vergessene Ganglinie.
	 */
	FineMaxAnzahlHistorischerGanglinienUeberschritten(Debug.FINE),

	/** Parameter: 1) Messquerschnitt, 2) Ereignistyp. */
	FineLerneEreignisTyp(Debug.FINE),

	/** Parameter: 1) Messquerschnitt, 2) Ergebnis des Pattern-Matching. */
	FineErgebnisPatternMatching(Debug.FINE),

	/**
	 * Parameter: 1) Messquerschnitt, 2) Index der Ganglinie, 3) die Ganglinie.
	 */
	FineAktualisiereGanglinie(Debug.FINE),

	/** Parameter: 1) Messquerschnitt, 2) Lernganglinie. */
	FineLerneGanglinie(Debug.FINE),

	/** Parameter: 1) neue letzte Lernzeit. */
	FineNeueLetzteLernzeit(Debug.FINE),

	/**
	 * Parameter: 1) aktuelle Zeit, 2) Begin des Lernzeitraums, 3) Ende des
	 * Lernzeitraums.
	 */
	FineInitialisiereLernen(Debug.FINE),

	/** Parameter: 1) Messquerschnitt. */
	FineSucheHistorischeGanglinieMitGeringstenAbstandNichtGefunden(Debug.FINE),

	/**
	 * Parameter: 1) Messquerschnitt, 2) Index der gefunden Ganglinie, 3)
	 * Abstand, 4) die gefundene Ganglinie.
	 */
	FineSucheHistorischeGanglinieMitGeringstenAbstandGefunden(Debug.FINE),

	/** Parameter: 1) Messquerschnitt, 2) Ereignistyp. */
	FinerSucheHistorischeGanglinieMitGeringstenAbstand(Debug.FINER),

	/**
	 * Parameter: 1) Messquerschnitt, 2) Index der aktuellen Ganglinie, 3)
	 * Abstand.
	 */
	FinerSucheHistorischeGanglinieMitGeringstenAbstandUntersuche(Debug.FINER),

	/** Parameter: 1) Messquerschnitt, 2) Ereignistyp. */
	FinerLerneRelativeGanglinieBezugsereignis(Debug.FINER),

	/** Parameter: 1) Messquerschnitt. */
	FinerLerneRelativeGanglinieKeinBezugsereignis(Debug.FINER),

	/** Parameter: 1) Messquerschnitt, 2) Ganglinie. */
	FinerGelernteGanglinie(Debug.FINER),

	/** Parameter: 1) Messquerschnitt, 2) Ganglinienliste. */
	FinestListeHistorischeGanglinien(Debug.FINEST),

	/** Parameter: 1) Messquerschnitt, 2) Ganglinienliste. */
	FinestGesicherteHistorischeGanglinien(Debug.FINEST),

	/** Parameter: 1) Messquerschnitt, 2) Archivganglinie. */
	FinerNormierteArchivganglinie(Debug.FINER);

	/** Das Resource-Bundle. */
	private static final ResourceBundle RESOURCE_BUNDLE = ResourceBundle
			.getBundle(GlLernenMsg.class.getCanonicalName());

	/** Die Eigenschaft {@code level}. */
	private final Level level;

	/**
	 * Initialisiert das Objekt.
	 *
	 * @param level
	 *            der Log-Level.
	 */
	private GlLernenMsg(final Level level) {
		this.level = level;
	}

	@Override
	public MessageGrade getBmvLevel() {
		if (level.equals(Debug.WARNING)) {
			return MessageGrade.WARNING;
		} else if (level.equals(Debug.ERROR)) {
			return MessageGrade.ERROR;
		}
		return null;

	}

	@Override
	public Level getLogLevel() {
		return level;
	}

	@Override
	public ResourceBundle getResourceBundle() {
		return RESOURCE_BUNDLE;
	}

	@Override
	public String toString() {
		return getResourceBundle().getString(name());
	}

}
